from decimal import Decimal

from transferwise.models.base import TFModel


class Quote(TFModel):
    _entity = 'quotes'
    profile = None  # Profile model
    source = None  # currency code
    target = None  # currency code
    rate_type = 'FIXED'
    target_amount = Decimal('0')
    type = None  # BALANCE_PAYOUT, BALANCE_CONVERSION

    def get_account_requirements(self, client):
        return client.get(f'{self._get_path(self.id)}/account-requirements',
                          additional_headers={'Accept-Minor-Version': '1'})



import base64
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError
from transferwise.client import config
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
import logging


class Client:
    def __init__(self, api_token, private_key=None, sandbox=True):
        self._api_token = api_token
        if private_key:
            self._private_key = serialization.load_pem_private_key(
                private_key,
                password=None,
                backend=default_backend()
            )
        self._transferwise_adapter = HTTPAdapter(max_retries=5)
        self.api_base_url = config.API_SANDBOX_URL if sandbox else config.API_PRODUCTION_URL
        self.api_version = config.API_VERSION
        self.session = requests.Session()
        self.session.mount(self.api_base_url, self._transferwise_adapter)
        self.logger = logging.getLogger(__name__)

    def get_url(self, path):
        return f'{self.api_base_url}/{path}'

    def get_headers(self):
        return {'Content-Type': 'application/json',
                'Authorization': f'Bearer {self._api_token}'}

    def get_approval_headers(self, approval_token):
        headers = self.get_headers()
        headers.update({
            'x-2fa-approval': approval_token,
            'X-Signature': self._get_approval_signature(approval_token)
        })
        return headers

    def _get_approval_signature(self, approval_token: str):
        signature = self._private_key.sign(
            approval_token,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        return base64.b64encode(signature).decode()

    def _request(self, url, method, headers, params=None, payload: dict = None, try_again=True):
        try:
            response = self.session.request(method, url, params=params, json=payload, headers=headers)
            approval_rejected = response.headers.get('x-2fa-approval-result', None) == 'REJECTED'
            if self._private_key and not response.ok and approval_rejected:
                if not try_again:
                    self.logger.warning(response.json())
                    return response.json()
                approval_token = response.headers.get('x-2fa-approval')
                approval_headers = self.get_approval_headers(approval_token)
                return self._request(url, method, approval_headers, params, payload, try_again=False)
            if not response.ok:
                self.logger.warning(response.json())
            return response.json()
        except ConnectionError as e:
            self.logger.warning(e)

    def get(self, path, additional_headers=None):
        api_url = self.get_url(path)
        headers = self.get_headers()
        if additional_headers:
            headers.update(additional_headers)
        return self._request(api_url, 'GET', headers)

    def post(self, path, payload: dict):
        api_url = self.get_url(path)
        headers = self.get_headers()
        return self._request(api_url, 'POST', headers, payload=payload)



from transferwise.models.base import TFModel


class Profile(TFModel):
    _entity = 'profiles'
    type = None  # personal, business
    details = None  # PersonalDetails, BusinessDetails


class PersonalDetails:
    first_name = None
    last_name = None
    date_of_birth = None
    phone_number = None
    avatar = None
    occupation = None
    primary_address = None  # Address model


class BusinessDetails:
    name = None
    registration_number = None
    acn = None
    abn = None
    arbn = None
    company_type = None
    company_role = None
    description_of_business = None
    primary_address = None  # Address model
    webpage = None



import re


def camelcase(string):
    string = re.sub(r"^[\-_\.]", '', str(string))
    if not string:
        return string
    return string[0].lower() + re.sub(r"[\-_\.\s]([a-z])", lambda matched: matched.group(1).upper(), string[1:])


def snakecase(string):
    string = re.sub(r"[\-\.\s]", '_', str(string))
    if not string:
        return string
    if string == string.upper():
        return string.lower()
    return string[0].lower() + re.sub(r"[A-Z]", lambda matched: '_' + matched.group(0).lower(), string[1:])



from transferwise.client.config import FUND_API_VERSION
from transferwise.models import Transfer, Profile
from transferwise.models.base import TFModel


class Fund(TFModel):
    _api_version = FUND_API_VERSION
    _api_path_pattern = '{api_version}/profiles/{profile_id}/transfers/{transfer_id}/{entity}'
    _entity = 'payments'
    profile = None  # Profile model
    transfer = None  # Transfer model



import base64
import unittest
import uuid
from cryptography.hazmat.backends.openssl.rsa import _RSAPrivateKey
from transferwise.client import Client
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding


class ClientTest(unittest.TestCase):
    def setUp(self) -> None:
        self.api_token = uuid.uuid4()
        self.key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        self.private_key = self.key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )
        self.pubkey = self.key.public_key()
        self.public_key = self.pubkey.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

    def test_client_without_private_key(self):
        client = Client(self.api_token)
        self.assertIsInstance(client, Client)

    def test_client_with_private_key(self):
        client = Client(self.api_token, self.private_key)
        self.assertIsInstance(client, Client)
        self.assertIsInstance(client._private_key, _RSAPrivateKey)

    def test_client_signs_approval_token(self):
        client = Client(self.api_token, self.private_key)
        approval_token = uuid.uuid4().bytes
        approval_headers = client.get_approval_headers(approval_token)
        self.assertEqual(approval_headers['x-2fa-approval'], approval_token)
        encoded_signature = approval_headers['X-Signature']
        signature = base64.b64decode(encoded_signature)
        self.pubkey.verify(
            signature,
            approval_token,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )

    def test_approval_headers_include_api_token(self):
        client = Client(self.api_token, self.private_key)
        approval_token = uuid.uuid4().bytes
        approval_headers = client.get_approval_headers(approval_token)
        self.assertEqual(approval_headers['Authorization'], f'Bearer {self.api_token}')



import re
from transferwise.models.base import TFModel
from transferwise.utils import snakecase


class Account(TFModel):
    _entity = 'accounts'
    currency = None  # currency code
    type = None  # sort_code, email
    profile = None  # Profile model
    account_holder_name = None
    owned_by_customer = False
    details = {}  # dict of actual recipient details according to country and currency

    def verify(self, account_requirements: list):
        errors = []
        for req in account_requirements:
            if req['type'] != self.type:
                continue
            for field in req['fields']:
                for group in field['group']:
                    self._validate_group(group, errors)
        return errors



import re
import uuid
from decimal import Decimal
from transferwise.client.config import API_VERSION
from transferwise.utils import camelcase, snakecase


class TFModel(object):
    _api_version = API_VERSION
    _entity = None  # must be set in concrete class
    _api_path_pattern = '{api_version}/{entity}'
    id = None

    def __init__(self):
        self._sync = {}

    def __getattribute__(self, item):
        if not item.startswith('_'):
            keys = set([k for k in self.__dict__.keys() if not k.startswith('_')])
            keys.update([k for k in self._sync.keys()])
            if item in keys and self._sync.get(item, None):
                return self._sync.get(item)

        return super().__getattribute__(item)


    @classmethod
    def create_from_dict(cls, data: dict):
        model = cls()
        for k, v in data.items():
            setattr(model, k, v)
        return model

    @classmethod
    def get(cls, client, object_id=None):
        response = client.get(cls._get_path(object_id))
        if isinstance(response, list):
            return [cls.create_from_dict(item) for item in response]
        return cls.create_from_dict(response)

    @classmethod
    def _get_path(cls, object_id=None, **kwargs):
        path = cls._api_path_pattern.format(api_version=cls._api_version, entity=cls._entity, **kwargs)
        if object_id:
            path += f'/{object_id}'
        return path

    def _validate_group(self, group, errors):
        local_key = snakecase(group['key'])
        if hasattr(self, local_key):
            value = getattr(self, local_key, None)
        else:
            value = self.details.get(group['key'], None)

        if group['required'] and not value:
            errors.append(group)
        elif not group['required'] and not value:
            return
        if group['validationRegexp']:
            match = re.match(r'^[A-Z]{2}[a-zA-Z0-9 ]{12,40}$', value)
            if not match:
                errors.append(group)
        if group['minLength'] and len(value) < group['minLength']:
            errors.append(group)
        if group['maxLength'] and len(value) > group['maxLength']:
            errors.append(group)
        if group['valuesAllowed']:
            if value not in [v['key'] for v in group['valuesAllowed']]:
                errors.append(group)

    def attr(self, key):
        return super().__getattribute__(key)

    def post(self, client, *args, **kwargs):
        data = client.post(self._get_path(**kwargs), payload=self.to_json())
        if isinstance(data, dict):
            for k, v in data.items():
                self._sync[snakecase(k)] = v
        elif isinstance(data, list):
            self._sync = data

    def to_json(self):
        serializable = {}
        attrs = dict(self.__class__.__dict__)
        attrs.update(self.__dict__)
        for k, v in attrs.items():
            if k.startswith('_') or callable(getattr(self, k)):
                continue
            k = camelcase(k)
            if isinstance(v, TFModel):
                serializable[k] = v.id  # if no id, it wasn't synced
            elif isinstance(v, Decimal):
                serializable[k] = float(f'{v.normalize():f}')
            elif isinstance(v, uuid.UUID):
                serializable[k] = f'{v}'
            else:
                serializable[k] = v
        return serializable




import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="transferwise-python-sdk",
    version="0.0.2",
    author="David Vartanian",
    author_email="davidvartanian@posteo.de",
    description="A light TransferWise SDK for Python",
    license="gpl-3.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/davidvartanian/transferwise-python-sdk",
    download_url="https://github.com/davidvartanian/transferwise-python-sdk/archive/v0.0.2.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=open("requirements.txt", "r").read().splitlines(),
    python_requires=">=3.6",
)



from transferwise.models.base import TFModel


class TransferRequirements(TFModel):
    """
    SEE https://api-docs.transferwise.com/#transfers-requirements
    """
    _entity = 'transfer-requirements'
    target_account = None  # Account model
    quote = None  # Quote model
    details = None  # dict with transfer details for transfer verification
    customer_transaction_id = None  # generated UUID for request idempotency


class Transfer(TFModel):
    _entity = 'transfers'
    target_account = None  # Account model
    quote = None  # Quote model
    customer_transaction_id = None  # generated UUID for request idempotency
    details = None  # dict with transfer details

    def verify(self, requirements: TransferRequirements):
        """
        Verify transfer data
        Raises an exception if verification didn't pass
        """
        errors = []
        for req in requirements:
            for field in req['fields']:
                for group in field['group']:
                    self._validate_group(group, errors)
        return errors



import random
import unittest
from transferwise.models import Account, Fund, Profile, Quote, Transfer


class AccountTest(unittest.TestCase):
    def setUp(self) -> None:
        self.account = Account()

    def test_account_instance_from_dict(self):
        self.assertIsInstance(Account.create_from_dict({'currency': 'EUR',
                                                        'type': 'BUSINESS'}), Account)

    def test_sync_data_accessible_as_attribute(self):
        self.account._sync['test_attr'] = 'value'
        self.assertEqual(self.account.test_attr, 'value')

    def test_to_json_is_serializable(self):
        self.account.profile = Profile.create_from_dict({'type': 'personal', 'details': {}})
        self.account.profile.id = random.randint(1, 100000)
        obj = self.account.to_json()
        self.assertEqual(obj['profile'], self.account.profile.id)
        self.assertIsInstance(obj, dict)


class FundTest(unittest.TestCase):
    def setUp(self) -> None:
        self.fund = Fund()

    def test_fund_instance_from_dict(self):
        self.assertIsInstance(Fund.create_from_dict({'currency': 'EUR',
                                                     'type': 'BUSINESS'}), Fund)

    def test_sync_data_accessible_as_attribute(self):
        self.fund._sync['test_attr'] = 'value'
        self.assertEqual(self.fund.test_attr, 'value')

    def test_to_json_is_serializable(self):
        self.fund.profile = Profile.create_from_dict({'type': 'personal', 'details': {}})
        self.fund.profile.id = random.randint(1, 100000)
        obj = self.fund.to_json()
        self.assertEqual(obj['profile'], self.fund.profile.id)
        self.assertIsInstance(obj, dict)


class ProfileTest(unittest.TestCase):
    def setUp(self) -> None:
        self.profile = Profile()

    def test_profile_instance_from_dict(self):
        self.assertIsInstance(Profile.create_from_dict({'currency': 'EUR',
                                                        'type': 'BUSINESS'}), Profile)

    def test_sync_data_accessible_as_attribute(self):
        self.profile._sync['test_attr'] = 'value'
        self.assertEqual(self.profile.test_attr, 'value')

    def test_to_json_is_serializable(self):
        obj = self.profile.to_json()
        self.assertIsInstance(obj, dict)


class QuoteTest(unittest.TestCase):
    def setUp(self) -> None:
        self.quote = Quote()

    def test_quote_instance_from_dict(self):
        self.assertIsInstance(Quote.create_from_dict({'currency': 'EUR',
                                                      'type': 'BUSINESS'}), Quote)

    def test_sync_data_accessible_as_attribute(self):
        self.quote._sync['test_attr'] = 'value'
        self.assertEqual(self.quote.test_attr, 'value')

    def test_to_json_is_serializable(self):
        self.quote.profile = Profile.create_from_dict({'type': 'personal', 'details': {}})
        self.quote.profile.id = random.randint(1, 100000)
        obj = self.quote.to_json()
        self.assertEqual(obj['profile'], self.quote.profile.id)
        self.assertIsInstance(obj, dict)


class TransferTest(unittest.TestCase):
    def setUp(self) -> None:
        self.transfer = Transfer()

    def test_transfer_instance_from_dict(self):
        self.assertIsInstance(Transfer.create_from_dict({'currency': 'EUR',
                                                         'type': 'BUSINESS'}), Transfer)

    def test_sync_data_accessible_as_attribute(self):
        self.transfer._sync['test_attr'] = 'value'
        self.assertEqual(self.transfer.test_attr, 'value')

    def test_to_json_is_serializable(self):
        self.transfer.account = Account.create_from_dict({'type': 'personal'})
        self.transfer.account.id = random.randint(1, 100000)
        obj = self.transfer.to_json()
        self.assertEqual(obj['account'], self.transfer.account.id)
        self.assertIsInstance(obj, dict)
